#!BPY

""" Registration info for Blender menus:
Name: 'Ldraw'
Blender: 233
Group: 'Import'
Tip: 'Import Ldraw (LEGO Cad) files.'
"""

#////////////////////////////////////// LDRAW ///////////////////////////////////*
#/////
#/////	SCRIPT: LDRAW_IMPORT.PY
#/////	VERSION: 0.01.00
#/////
#/////	AUTHORS:	Alessandro V. 2004 - alv67@katamail.com
#/////
#/////	REQUIRMENTS:	Blender 2.33 (don't know if it works in version 2.32)
#/////
#/////  REFERENCES:     LDraw.org Official Specifications
#/////                  - LDraw File Format          (May 29, 2003)
#/////                  - MPD Language Extension     (May 29, 2003)
#/////                  - LDraw Language Extension for Clipping
#/////                                               (Sep 22, 2003)
#////                   - LDraw Color Chart          (Sep 05, 2003)
#/////
#/////	DONE:	        Read single model DAT, LDR files
#/////                  Skip lines "0" comment or meta-commands
#/////                  Import     'Triangles' ("3") and
#/////                        'Quadrilaterals' ("4") as Mesh Object
#/////                  Import     'Part-file' ("1") as linked Object
#/////
#/////	TODO:	        - Correct face normals
#/////                  - Import       "Line"  ("2") command (if needed)
#/////                      "Conditiona-line"  ("5") command (if needed)
#/////                  - Set part color creating a complete color pallet
#/////                    with transparencies, complementar color
#/////                  - Import some meta commands :
#/////                      "0 BFC [CLIP|NOCLIP] [CW|CCW] [INVERTNEXT]"
#/////                  - Read Multi-Part DAT files (MPD extension)
#/////                      "0 FILE model.ldr"  Multi-part DAT or MPD
#/////                  - Create a single object for any library part
#/////                    and use linked duplicates in the model
#/////
#/////	LICENSE:        The script is under GPL and feel free to use it.
#/////
#/////
#/////- REVISION HISTORY --------------------------------------------------------
#/////
#/////  Version 0.01.00 - 29 june 2004
#/////  ------------------------------
#/////  Initial release
#/////
#////////////////////////////////////// LDRAW ///////////////////////////////////*


import Blender
from Blender import NMesh, Material, Texture, Scene, Lamp, Image, sys, Mathutils
from Blender.Draw import *
from Blender.BGL import *

LDrawPath="C:\Ldraw"      # LDRAW base installation directory

EVENT_PATHCHANGE=		1
EVENT_IMPORT=			2
EVENT_IMPORT_MATS=		3
EVENT_QUIT=			4


#//////////////////////////////// COLOR TABLE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
#/////
#/////  Values from Official "Ldraw Color Chart"
#/////  List of [Red, Green, Blu, Alpha]
#/////
LdrawColor = [ [  51,  51,  51, 255],    #   0  Black LDraw
               [   0,  51, 178, 255],    #   1  Blue LDraw
               [   0, 127,  51, 255],    #   2  Green LDraw
               [   0, 153, 159, 255],    #   3  Dark Cyan LDraw
               [ 204,   0,   0, 255],    #   4  Red LDraw
               [ 223, 102, 149, 255],    #   5  Magenta LDraw
               [ 102,  51,   0, 255],    #   6  Brown LDraw
               [ 153, 153, 153, 255],    #   7  Grey LDraw
               [ 102, 102,  88, 255],    #   8  Dark Grey LDraw
               [   0, 128, 255, 255],    #   9  Light Blue LDraw
               [  51, 255, 102, 255],    #  10  Light Green LDraw
               [  51, 166, 167, 255],    #  11  Cyan LDraw
               [ 255, 201, 196, 255],    #  12  Light Red LDraw
               [ 255, 176, 204, 255],    #  13  Pink LDraw
               [ 255, 229,   0, 255],    #  14  Yellow LDraw
               [ 255, 255, 255, 255]]    #  15  White LDraw
#/////
#/////  Color   16 = Null-colour
#/////        Pieces using null colour code use whatever colour code is 
#/////        entered on the command line when rendering with LDraw or,
#/////        if no colour code is specified, they default to a #7 Grey.
#/////  Clear Colours.
#/////        Add 32 to get clear versions of the colours (Alpha = 200).
#/////        For example, to get clear red use 36 (32 for clear + 4 for red)
#/////
#//////////////////////////////// COLOR TABLE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*



#//////////////////////////////////// ADDPART \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
#/////
#/////	Extract from the string 'aLine' object position, rotation and
#/////  scaling, import file stated in the line and fill object 'aObj'
#/////  with these informations
#///// In LDRAW coordinate system is different from Blender's.
#/////    Positive axis vectors points to:
#/////    LDRAW axis   : X->Right,  Y->Down,    Z->Forward
#/////    BLENDER axis : X->Right,  Y->Forward, Z->Up
#/////
#//////////////////////////////////// ADDPART \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
def AddPart( aObj, aLine, aPath=""):
    TranfMatrix = []            # 3D Transformation Matrix of the object
    color = CValue()
    vertex = CVertex()
    x1 = CValue(); x2 = CValue(); x3 = CValue();
    y1 = CValue(); y2 = CValue(); y3 = CValue();
    z1 = CValue(); z2 = CValue(); z3 = CValue();
    if TrimL(aLine)[0] != '1':  # Check if line contains correct command
        return
    aLine = GetValue(aLine[1:], color)
    aLine = GetVertex(aLine, vertex)
#    aLine = GetValue(aLine, x1)
#    aLine = GetValue(aLine, x2)
#    aLine = GetValue(aLine, x3)
#    aLine = GetValue(aLine, y1)
#    aLine = GetValue(aLine, y2)
#    aLine = GetValue(aLine, y3)
#    aLine = GetValue(aLine, z1)
#    aLine = GetValue(aLine, z2)
#    aLine = GetValue(aLine, z3)
    aLine = GetValue(aLine, x1)
    aLine = GetValue(aLine, x3)
    aLine = GetValue(aLine, x2)
    aLine = GetValue(aLine, z1)
    aLine = GetValue(aLine, z3)
    aLine = GetValue(aLine, z2)
    aLine = GetValue(aLine, y1)
    aLine = GetValue(aLine, y3)
    aLine = GetValue(aLine, y2)
    PartName = TrimR(aLine)

    # Search for part in current import directory
    strFileName = aPath + sys.sep + PartName
    try:
        aFile = open(strFileName, "rb")
    except IOError:
        # Search for part in Ldraw PARTS directory
        strFileName = LDrawPath + sys.sep + "PARTS" + sys.sep + PartName
        try:
            aFile = open(strFileName, "rb")
        except IOError:
            # Search for part in Ldraw P directory
            strFileName = LDrawPath + sys.sep + "P" + sys.sep + PartName
            try:
                aFile = open(strFileName, "rb")
            except IOError:
                print("Unable to find part file %s!" % (PartName))
                return
    aFile.close()

#DBG    print "Loading Part ",strFileName
    ImportDAT(strFileName, aObj)
    TranfMatrix = Blender.Mathutils.Matrix([x1.Value,y1.Value,z1.Value,0],
                                           [x2.Value,y2.Value,z2.Value,0],
                                           [x3.Value,y3.Value,z3.Value,0],
                                           [vertex.x,vertex.y,vertex.z,1])
#    print matrice
    sc = Blender.Scene.getCurrent ()
    sc.link(aObj)
    aObj.setMatrix(TranfMatrix)
#    Blender.Redraw()


#////////////////////////////////// ADDTRIANGLE \\\\\\\\\\\\\\\\\\\\\\\\\\\\*
#/////
#/////	Extract from the string 'aLine' vertex information of the
#/////    Triangle primitive, create the face and add it to the Mesh
#/////    object passed 'aMesh'
#/////
#////////////////////////////////// ADDTRIANGLE \\\\\\\\\\\\\\\\\\\\\\\\\\\\*
def AddTriangle( aMesh, aLine):
    idx = 0
    vertex = []; color = []; v = []; f = []
    if aLine[0] != '3':
        return
    idx = len(aMesh.verts)
    vertex = CVertex()
    color  = CValue()
    aLine = GetValue(aLine[1:], color)
    aLine = GetVertex(aLine, vertex)
    v = NMesh.Vert(vertex.x, vertex.y, vertex.z)
    aMesh.verts.append(v)
    aLine = GetVertex(aLine, vertex)
    v = NMesh.Vert(vertex.x, vertex.y, vertex.z)
    aMesh.verts.append(v)
    aLine = GetVertex(aLine, vertex)
    v = NMesh.Vert(vertex.x, vertex.y, vertex.z)
    aMesh.verts.append(v)
    # Create Face
    f = NMesh.Face()
    f.v.append(aMesh.verts[idx])
    f.v.append(aMesh.verts[idx+1])
    f.v.append(aMesh.verts[idx+2])
    aMesh.faces.append(f)


#/////////////////////////////// ADDQUADRILATERAL \\\\\\\\\\\\\\\\\\\\\\\\\\*
#/////
#/////	Extract from the string 'aLine' vertex information of the
#/////    Quadrilateral primitive, create the face and add it to the Mesh
#/////    object passed 'aMesh'
#/////
#/////////////////////////////// ADDQUADRILATERAL \\\\\\\\\\\\\\\\\\\\\\\\\\*
def AddQuadrilateral( aMesh, aLine):
    idx = 0
    vertex = []; color = []; v = []; f = []
    if aLine[0] != '4':
        return
    idx = len(aMesh.verts)
    vertex = CVertex()
    color  = CValue()
    aLine = GetValue(aLine[1:], color)
    aLine = GetVertex(aLine, vertex)
    v = NMesh.Vert(vertex.x, vertex.y, vertex.z)
    aMesh.verts.append(v)
    aLine = GetVertex(aLine, vertex)
    v = NMesh.Vert(vertex.x, vertex.y, vertex.z)
    aMesh.verts.append(v)
    aLine = GetVertex(aLine, vertex)
    v = NMesh.Vert(vertex.x, vertex.y, vertex.z)
    aMesh.verts.append(v)
    aLine = GetVertex(aLine, vertex)
    v = NMesh.Vert(vertex.x, vertex.y, vertex.z)
    aMesh.verts.append(v)
    # Create Face
    f = NMesh.Face()
    f.v.append(aMesh.verts[idx])
    f.v.append(aMesh.verts[idx+1])
    f.v.append(aMesh.verts[idx+2])
    f.v.append(aMesh.verts[idx+3])
    aMesh.faces.append(f)


#////////////////////////////////// IMPORTDAT \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
#/////
#/////	Main function of the class, open file and import the LDraw File
#/////
#////////////////////////////////// IMPORTDAT \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
def ImportDAT( strFileName , aObj ):
    aMesh = []
    mFile = []    # File pointer
    Line = ""
    ChildList = []     # Child of aObj list
    aPath = sys.dirname(strFileName)
    aName = uppercase(sys.basename(strFileName))


    # Open the DAT file
    try:
        mFile = open(strFileName, "rb")
    except IOError:
        # Make sure we have a valid file pointer (we found the file)
        print("Unable to find the file %s!" % (strFileName))
        return

    aMesh = Blender.NMesh.New(aName)

    aObj.setName(aName)
#DBG    print "Creating Object: ", aName

    # Once the file is opened, we read all the lines (DAT file is a simle
    # text file
    Line = mFile.readline()
    while Line <> "":
         Line = TrimL(Line)
         if Line=="":
             pass
         elif Line[0]=='0':  # Remark
             pass
         elif Line[0]=='1':  # Part File
             bObj = Blender.Object.New('Mesh')
             AddPart(bObj, Line, aPath)
             ChildList = ChildList + [bObj]
             aObj.makeParent(ChildList)
         elif Line[0]=='2':  # Line
             pass
         elif Line[0]=='3':  # Triangle
             AddTriangle(aMesh, Line)
         elif Line[0]=='4':  # Quadrilateral
             AddQuadrilateral(aMesh, Line)
         elif Line[0]=='5':  # Optional-Line
             pass
         else :
             pass
         Line = mFile.readline()

    mFile.close()
    # If the mesh contains vertex then link this mesh in the object
    if aMesh.verts:
       aObj.link(aMesh)

    return


#/////////////////////////////////// CVERTEX ///////////////////////////////////*
#/////
#///// This is the 3d vertex class
#/////
#/////////////////////////////////// CVERTEX ///////////////////////////////////*
class CVertex(object):
      x = 0.0
      y = 0.0
      z = 0.0


#//////////////////////////////////// CVALUE ///////////////////////////////////*
#/////
#///// This is the float number class. Used in functions to read a float number
#/////
#//////////////////////////////////// CVALUE ///////////////////////////////////*
class CValue(object):
      Value = 0.0


#//////////////////////////////////// TRIML /////////////////////////////////////*
#/////
#///// This function erase spaces and tabs from head (left) of the string
#/////
#//////////////////////////////////// TRIML /////////////////////////////////////*
def TrimL(s):
    i = 0
    while i < len(s):
        if s[i]!=' ' and s[i]!='\t':
           return s[i:]
        i = i + 1
    return s


#//////////////////////////////////// TRIMR /////////////////////////////////////*
#/////
#///// This function erase spaces,tabs,LF and CR from tail (right) of the string
#/////
#//////////////////////////////////// TRIMR /////////////////////////////////////*
def TrimR(s):
    i = len(s)-1
    while i > 0:
        if s[i]!=' ' and s[i]!='\t' and s[i]!='\n' and s[i]!='\r':
           return s[0:i+1]
        i = i - 1
    return s



#////////////////////////////////// LOWERCASE //////////////////////////////////*
#/////
#///// This function set all character of the string into lowercase chars
#/////
#///////////////////////////////// LOWERCASE ///////////////////////////////////*
def lowercase(s):
    ret = ""
    i = 0
    while i < len(s):
        if s[i] >= 'A' and s[i] <= 'Z':
            ret = ret + chr(ord(s[i])+ord('a')-ord('A'))
        else :
            ret = ret + s[i]
        i = i + 1
    return ret


#////////////////////////////////// UPPERCASE //////////////////////////////////*
#/////
#///// This function set all character of the string into uppercase chars
#/////
#///////////////////////////////// UPPERCASE ///////////////////////////////////*
def uppercase(s):
    ret = ""
    i = 0
    while i < len(s):
        if s[i] >= 'a' and s[i] <= 'z':
            ret = ret + chr(ord(s[i])+ord('A')-ord('a'))
        else :
            ret = ret + s[i]
        i = i + 1
    return ret


#//////////////////////////////////// GETVALUE ///////////////////////////////////*
#/////
#///// Extract first number from the string 's', writes the number in 'value'
#///// object and return string whithout first number
#/////
#//////////////////////////////////// GETVALUE ///////////////////////////////////*
def GetValue(s, value):
    i = 0
    s = TrimL(s)
    while i < len(s):
        if s[i]==' ' or s[i]=='\n' or s[i]=='\r':
            value.Value = float(s[0:i])
            break
        i = i + 1
    return TrimL(s[i:])


#//////////////////////////////////// GETVERTEX /////////////////////////////////*
#/////
#///// Extract three numbers from the string 's', fill the 'vertex' object
#///// object and return string whithout tis numbers
#///// In LDRAW coordinate system is different from Blender's.
#/////    Positive axis vectors points to:
#/////    LDRAW axis   : X->Right,  Y->Down,    Z->Forward
#/////    BLENDER axis : X->Right,  Y->Forward, Z->Up
#/////
#//////////////////////////////////// GETVERTEX /////////////////////////////////*
def GetVertex(s, vertex):
    i = 0
    v = CValue()
    s = GetValue(s,v)
    vertex.x = v.Value
    s = GetValue(s,v)
    vertex.z = -v.Value
    s = GetValue(s,v)
    vertex.y = v.Value
    return TrimL(s)


#//////////////////////////////////// EXITGUI /////////////////////////////////*
#/////
#/////	Exit form the gui
#/////
#//////////////////////////////////// EXITGUI /////////////////////////////////*
def ExitGUI ():
        Exit()

#//////////////////////////////////// EVENTGUI /////////////////////////////////*
#/////
#/////	Process events
#/////
#//////////////////////////////////// EVENTGUI /////////////////////////////////*
def	EventGUI (event):
        global bMaterials
	if (event==EVENT_IMPORT):
                bMaterials = False
                Blender.Window.FileSelector(Import, "LDRAW Import")
	elif (event==EVENT_QUIT):
                ExitGUI()

#//////////////////////////////////// DRAWGUI /////////////////////////////////*
#/////
#/////	Draw the gui in blender
#/////
#//////////////////////////////////// DRAWGUI /////////////////////////////////*
def DrawGUI():

	glClearColor(0.6,0.6,0.6,0)
	glClear(GL_COLOR_BUFFER_BIT)
	glRasterPos2d(100,64)
	Text("Import LDRAW (Lego CAD) file")
	Button ("Import",EVENT_IMPORT,32,10,100,32,"Select File to Import")
	Button ("Cancel",EVENT_QUIT,252,10,100,32,"Do not Import")


#//////////////////////////////////// IMPORT /////////////////////////////////*
#/////
#/////	Load the 3ds file, parse it and import everything in blender
#/////
#//////////////////////////////////// IMPORT /////////////////////////////////*
def Import(FileName):
#DBG        print "--------------- START IMPORT ----------------"
        scene = Blender.Scene.getCurrent ()
        aObj = Blender.Object.New('Mesh')
	try:
		ImportDAT(FileName, aObj);
	except IOError:
		Draw()
		print "error while loading Ldraw file make sure path is correct"

        scene.link (aObj)
        aObj.setLocation(0,0,0)
        Blender.Redraw()


#//////////////////////////////////// REGISTERGUI /////////////////////////////////*
#/////
#/////   Register callbacks for windowing
#/////
#/////   DrawGui  : A function to draw the screen, taking no arguments
#/////   None     : No function to handle keyboard and mouse input events
#/////   EventGUI : A function to handle Draw Button events, taking 'event' argument
#/////
#//////////////////////////////////// REGISTERGUI /////////////////////////////////*
def RegisterGUI ():
	Register (DrawGUI,None,EventGUI)

#////////////////////////////////////  MAIN SCRIPT /////////////////////////////////*
#/////
#/////	Start the script
#/////
#////////////////////////////////////  MAIN SCRIPT /////////////////////////////////*
RegisterGUI()

